@@ -3,9 +3,9 @@
  *
  * Common functions for SAKURA and TOMOYO.
  *
- * Copyright (C) 2005-2006  NTT DATA CORPORATION
+ * Copyright (C) 2005-2007  NTT DATA CORPORATION
  *
- * Version: 1.3.1   2006/12/08
+ * Version: 1.4   2007/04/01
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -34,6 +34,22 @@ static const int lookup_flags = LOOKUP_FOLLOW | LOOKUP_POSITIVE;
 #include <linux/ccs_proc.h>
 #include <linux/tomoyo.h>
 
+#ifdef CONFIG_TOMOYO_MAX_ACCEPT_ENTRY
+#define MAX_ACCEPT_ENTRY (CONFIG_TOMOYO_MAX_ACCEPT_ENTRY)
+#else
+#define MAX_ACCEPT_ENTRY 2048
+#endif
+#ifdef CONFIG_TOMOYO_MAX_GRANT_LOG
+#define MAX_GRANT_LOG (CONFIG_TOMOYO_MAX_GRANT_LOG)
+#else
+#define MAX_GRANT_LOG 1024
+#endif
+#ifdef CONFIG_TOMOYO_MAX_REJECT_LOG
+#define MAX_REJECT_LOG (CONFIG_TOMOYO_MAX_REJECT_LOG)
+#else
+#define MAX_REJECT_LOG 1024
+#endif
+
 /*************************  VARIABLES  *************************/
 
 /* /sbin/init started? */
@@ -50,17 +66,14 @@ static struct {
 	[CCS_TOMOYO_MAC_FOR_FILE]        = { "MAC_FOR_FILE",        0, 3 },
 	[CCS_TOMOYO_MAC_FOR_ARGV0]       = { "MAC_FOR_ARGV0",       0, 3 },
 	[CCS_TOMOYO_MAC_FOR_NETWORK]     = { "MAC_FOR_NETWORK",     0, 3 },
-	[CCS_TOMOYO_MAC_FOR_BINDPORT]    = { "MAC_FOR_BINDPORT",    0, 3 },
-	[CCS_TOMOYO_MAC_FOR_CONNECTPORT] = { "MAC_FOR_CONNECTPORT", 0, 3 },
 	[CCS_TOMOYO_MAC_FOR_SIGNAL]      = { "MAC_FOR_SIGNAL",      0, 3 },
 	[CCS_SAKURA_DENY_CONCEAL_MOUNT]  = { "DENY_CONCEAL_MOUNT",  0, 3 },
 	[CCS_SAKURA_RESTRICT_CHROOT]     = { "RESTRICT_CHROOT",     0, 3 },
 	[CCS_SAKURA_RESTRICT_MOUNT]      = { "RESTRICT_MOUNT",      0, 3 },
 	[CCS_SAKURA_RESTRICT_UNMOUNT]    = { "RESTRICT_UNMOUNT",    0, 3 },
-	[CCS_SAKURA_DENY_PIVOT_ROOT]     = { "DENY_PIVOT_ROOT",     0, 3 },
-	[CCS_SAKURA_TRACE_READONLY]      = { "TRACE_READONLY",      0, 1 },
+	[CCS_SAKURA_RESTRICT_PIVOT_ROOT] = { "RESTRICT_PIVOT_ROOT", 0, 3 },
 	[CCS_SAKURA_RESTRICT_AUTOBIND]   = { "RESTRICT_AUTOBIND",   0, 1 },
-	[CCS_TOMOYO_MAX_ACCEPT_FILES]    = { "MAX_ACCEPT_FILES",    MAX_ACCEPT_FILES, INT_MAX },
+	[CCS_TOMOYO_MAX_ACCEPT_ENTRY]    = { "MAX_ACCEPT_ENTRY",    MAX_ACCEPT_ENTRY, INT_MAX },
 	[CCS_TOMOYO_MAX_GRANT_LOG]       = { "MAX_GRANT_LOG",       MAX_GRANT_LOG, INT_MAX },
 	[CCS_TOMOYO_MAX_REJECT_LOG]      = { "MAX_REJECT_LOG",      MAX_REJECT_LOG, INT_MAX },
 	[CCS_TOMOYO_VERBOSE]             = { "TOMOYO_VERBOSE",      1, 1 },
@@ -69,7 +82,7 @@ static struct {
 
 typedef struct {
 	unsigned int value[CCS_MAX_CONTROL_INDEX];
-	const char *comment;
+	const struct path_info *comment;
 } PROFILE;
 
 static PROFILE *profile_ptr[MAX_PROFILES];
@@ -92,14 +105,6 @@ static int isRoot(void)
 	return !current->uid && !current->euid;
 }
 
-int strendswith(const char *name, const char *tail)
-{
-	int len;
-	if (!name || !tail) return 0;
-	len = strlen(name) - strlen(tail);
-	return len >= 0 && strcmp(name + len, tail) == 0;
-}
-
 /*
  * Format string.
  * Leading and trailing whitespaces are removed.
@@ -184,7 +189,51 @@ int IsCorrectPath(const char *filename, const int start_type, const int pattern_
 	return 0;
 }
 
-int PathDepth(const char *pathname)
+/*
+ *  Check whether the given domainname follows the naming rules.
+ *  Returns nonzero if follows, zero otherwise.
+ */
+int IsCorrectDomain(const unsigned char *domainname, const char *function)
+{
+	unsigned char c, d, e;
+	const char *org_domainname = domainname;
+	if (!domainname || strncmp(domainname, ROOT_NAME, ROOT_NAME_LEN)) goto out;
+	domainname += ROOT_NAME_LEN;
+	if (!*domainname) return 1;
+	do {
+		if (*domainname++ != ' ') goto out;
+		if (*domainname++ != '/') goto out;
+		while ((c = *domainname) != '\0' && c != ' ') {
+			domainname++;
+			if (c == '\\') {
+				switch ((c = *domainname++)) {
+				case '\\':  /* "\\" */
+					continue;
+				case '0':   /* "\ooo" */
+				case '1':
+				case '2':
+				case '3':
+					if ((d = *domainname++) >= '0' && d <= '7' && (e = *domainname++) >= '0' && e <= '7') {
+						const unsigned char f =
+							(((unsigned char) (c - '0')) << 6) +
+							(((unsigned char) (d - '0')) << 3) +
+							(((unsigned char) (e - '0')));
+						if (f && (f <= ' ' || f >= 127)) continue; /* pattern is not \000 */
+					}
+				}
+				goto out;
+			} else if (c < ' ' || c >= 127) {
+				goto out;
+			}
+		}
+	} while (*domainname);
+	return 1;
+ out:
+	printk(KERN_DEBUG "%s: Invalid domainname '%s'\n", function, org_domainname);
+	return 0;
+}
+
+static int PathDepth(const char *pathname)
 {
 	int i = 0;
 	if (pathname) {
@@ -197,6 +246,40 @@ int PathDepth(const char *pathname)
 	return i;
 }
 
+static int const_part_length(const char *filename)
+{
+	int len = 0;
+	if (filename) {
+		char c;
+		while ((c = *filename++) != '\0') {
+			if (c != '\\') { len++; continue; }
+			switch (c = *filename++) {
+			case '\\':  /* "\\" */
+				len += 2; continue;
+			case '0':   /* "\ooo" */
+			case '1':
+			case '2':
+			case '3':
+				if ((c = *filename++) >= '0' && c <= '7' && (c = *filename++) >= '0' && c <= '7') { len += 4; continue; }
+			}
+			break;
+		}
+	}
+	return len;
+}
+
+void fill_path_info(struct path_info *ptr)
+{
+	const char *name = ptr->name;
+	const int len = strlen(name);
+	ptr->total_len = len;
+	ptr->const_len = const_part_length(name);
+	ptr->is_dir = len && (name[len - 1] == '/');
+	ptr->is_patterned = (ptr->const_len < len);
+	ptr->hash = full_name_hash(name, len);
+	ptr->depth = PathDepth(name);
+}
+
 static int FileMatchesToPattern(const char *filename, const char *filename_end, const char *pattern, const char *pattern_end)
 {
 	while (filename < filename_end && pattern < pattern_end) {
@@ -305,12 +388,15 @@ static int FileMatchesToPattern(const char *filename, const char *filename_end,
  *    \a     1 alphabet character.
  */
 
-int PathMatchesToPattern(const char *pathname, const char *pattern)
+int PathMatchesToPattern(const struct path_info *pathname0, const struct path_info *pattern0)
 {
-	if (!pathname || !pattern) return 0;
-	/* if pattern doesn't contain '\', I can use strcmp(). */
-	if (!strchr(pattern, '\\')) return !strcmp(pathname, pattern);
-	if (PathDepth(pathname) != PathDepth(pattern)) return 0;
+	//if (!pathname || !pattern) return 0;
+	const char *pathname = pathname0->name, *pattern = pattern0->name;
+	const int len = pattern0->const_len;
+	if (!pattern0->is_patterned) return !pathcmp(pathname0, pattern0);
+	if (pathname0->depth != pattern0->depth) return 0;
+	if (strncmp(pathname, pattern, len)) return 0;
+	pathname += len; pattern += len;
 	while (*pathname && *pattern) {
 		const char *pathname_delimiter = strchr(pathname, '/'), *pattern_delimiter = strchr(pattern, '/');
 		if (!pathname_delimiter) pathname_delimiter = strchr(pathname, '\0');
@@ -351,10 +437,7 @@ const char *GetEXE(void)
 		struct vm_area_struct *vma = current->mm->mmap;
 		while (vma) {
 			if ((vma->vm_flags & VM_EXECUTABLE) && vma->vm_file) {
-				char *buf = ccs_alloc(PAGE_SIZE);
-				if (buf == NULL) return NULL;
-				if (realpath_from_dentry(vma->vm_file->f_dentry, vma->vm_file->f_vfsmnt, buf, PAGE_SIZE - 1) == 0) return (const char *) buf;
-				ccs_free(buf); return NULL;
+				return realpath_from_dentry(vma->vm_file->f_dentry, vma->vm_file->f_vfsmnt);
 			}
 			vma = vma->vm_next;
 		}
@@ -373,7 +456,11 @@ const char *GetMSG(const int is_enforce)
 unsigned int CheckCCSFlags(const unsigned int index)
 {
 	const u8 profile = current->domain_info->profile;
-	return sbin_init_started && index < CCS_MAX_CONTROL_INDEX && profile < MAX_PROFILES && profile_ptr[profile] ? profile_ptr[profile]->value[index] : 0;
+	return sbin_init_started && index < CCS_MAX_CONTROL_INDEX
+#if MAX_PROFILES != 256
+		&& profile < MAX_PROFILES
+#endif
+		&& profile_ptr[profile] ? profile_ptr[profile]->value[index] : 0;
 }
 
 unsigned int TomoyoVerboseMode(void)
@@ -393,21 +480,6 @@ unsigned int CheckCCSAccept(const unsigned int index)
 	return CheckCCSFlags(index) == 1;
 }
 
-unsigned int GetMaxAutoAppendFiles(void)
-{
-	return CheckCCSFlags(CCS_TOMOYO_MAX_ACCEPT_FILES);
-}
-
-unsigned int GetMaxGrantLog(void)
-{
-	return CheckCCSFlags(CCS_TOMOYO_MAX_GRANT_LOG);
-}
-
-unsigned int GetMaxRejectLog(void)
-{
-	return CheckCCSFlags(CCS_TOMOYO_MAX_REJECT_LOG);
-}
-
 static PROFILE *FindOrAssignNewProfile(const unsigned int profile)
 {
 	static DECLARE_MUTEX(profile_lock);
@@ -480,15 +552,10 @@ static int ReadStatus(IO_BUFFER *head)
 				case -1: // Dummy
 #ifndef CONFIG_TOMOYO_MAC_FOR_FILE
 				case CCS_TOMOYO_MAC_FOR_FILE:
-				case CCS_TOMOYO_MAX_ACCEPT_FILES:
 #endif
 #ifndef CONFIG_TOMOYO_MAC_FOR_ARGV0
 				case CCS_TOMOYO_MAC_FOR_ARGV0:
 #endif
-#ifndef CONFIG_TOMOYO_MAC_FOR_NETWORKPORT
-				case CCS_TOMOYO_MAC_FOR_BINDPORT:
-				case CCS_TOMOYO_MAC_FOR_CONNECTPORT:
-#endif
 #ifndef CONFIG_TOMOYO_MAC_FOR_NETWORK
 				case CCS_TOMOYO_MAC_FOR_NETWORK:
 #endif
@@ -507,16 +574,14 @@ static int ReadStatus(IO_BUFFER *head)
 #ifndef CONFIG_SAKURA_RESTRICT_UNMOUNT
 				case CCS_SAKURA_RESTRICT_UNMOUNT:
 #endif
-#ifndef CONFIG_SAKURA_DENY_PIVOT_ROOT
-				case CCS_SAKURA_DENY_PIVOT_ROOT:
-#endif
-#ifndef CONFIG_SAKURA_TRACE_READONLY
-				case CCS_SAKURA_TRACE_READONLY:
+#ifndef CONFIG_SAKURA_RESTRICT_PIVOT_ROOT
+				case CCS_SAKURA_RESTRICT_PIVOT_ROOT:
 #endif
 #ifndef CONFIG_SAKURA_RESTRICT_AUTOBIND
 				case CCS_SAKURA_RESTRICT_AUTOBIND:
 #endif
 #ifndef CONFIG_TOMOYO
+				case CCS_TOMOYO_MAX_ACCEPT_ENTRY:
 				case CCS_TOMOYO_MAX_GRANT_LOG:
 				case CCS_TOMOYO_MAX_REJECT_LOG:
 				case CCS_TOMOYO_VERBOSE:
@@ -524,9 +589,7 @@ static int ReadStatus(IO_BUFFER *head)
 					continue;
 				}
 				if (j == CCS_PROFILE_COMMENT) {
-					const char *comment = profile->comment;
-					if (!comment) comment = "";
-					if (io_printf(head, "%u-%s=%s\n", i, ccs_control_array[CCS_PROFILE_COMMENT].keyword, comment)) break;
+					if (io_printf(head, "%u-%s=%s\n", i, ccs_control_array[CCS_PROFILE_COMMENT].keyword, profile->comment ? profile->comment->name : "")) break;
 				} else {
 					if (io_printf(head, "%u-%s=%u\n", i, ccs_control_array[j].keyword, profile->value[j])) break;
 				}
@@ -550,31 +613,67 @@ static int ReadStatus(IO_BUFFER *head)
 /*************************  POLICY MANAGER HANDLER  *************************/
 
 typedef struct policy_manager_entry {
-	struct policy_manager_entry *next; /* Pointer to next record. NULL if none. */
-	const char *exe;                   /* Filename. Never NULL.                 */
+	struct policy_manager_entry *next;
+	const struct path_info *manager;
+	u8 is_domain;
+	u8 is_deleted;
 } POLICY_MANAGER_ENTRY;
 
-static POLICY_MANAGER_ENTRY policy_manager_list = { NULL, "" };
+static POLICY_MANAGER_ENTRY *policy_manager_list = NULL;
+
+static int AddManagerEntry(const char *manager, u8 is_delete)
+{
+	POLICY_MANAGER_ENTRY *new_entry, *ptr;
+	static DECLARE_MUTEX(lock);
+	const struct path_info *saved_manager;
+	int error = -ENOMEM;
+	u8 is_domain = 0;
+	if (!isRoot()) return -EPERM;
+	if (IsDomainDef(manager)) {
+		if (!IsCorrectDomain(manager, __FUNCTION__)) return -EINVAL;
+		is_domain = 1;
+	} else {
+		if (!IsCorrectPath(manager, 1, -1, -1, __FUNCTION__)) return -EINVAL;
+	}
+	if ((saved_manager = SaveName(manager)) == NULL) return -ENOMEM;
+	down(&lock);
+	for (ptr = policy_manager_list; ptr; ptr = ptr->next) {
+		if (ptr->manager == saved_manager) {
+			ptr->is_deleted = is_delete;
+			error = 0;
+			goto out;
+		}
+	}
+	if (is_delete) {
+		error = -ENOENT;
+		goto out;
+	}
+	if ((new_entry = (POLICY_MANAGER_ENTRY *) alloc_element(sizeof(POLICY_MANAGER_ENTRY))) == NULL) goto out;
+	new_entry->manager = saved_manager;
+	new_entry->is_domain = is_domain;
+	mb(); /* Instead of using spinlock. */
+	if ((ptr = policy_manager_list) != NULL) {
+		while (ptr->next) ptr = ptr->next; ptr->next = new_entry;
+	} else {
+		policy_manager_list = new_entry;
+	}
+	error = 0;
+ out:
+	up(&lock);
+	if (!error) UpdateCounter(CCS_UPDATES_COUNTER_MANAGER);
+	return error;
+}
 
 static int AddManagerPolicy(IO_BUFFER *head)
 {
-	const char *exe = head->write_buf;
-	POLICY_MANAGER_ENTRY *new_entry, *ptr;
-	static spinlock_t lock = SPIN_LOCK_UNLOCKED;
-	const char *saved_exe;
+	const char *data = head->write_buf;
+	int is_delete = 0;
 	if (!isRoot()) return -EPERM;
-	if (!IsCorrectPath(exe, 1, -1, -1, __FUNCTION__)) return -EINVAL;
-	UpdateCounter(CCS_UPDATES_COUNTER_MANAGER);
-	/* I don't want to add if it was already added. */
-	for (ptr = policy_manager_list.next; ptr; ptr = ptr->next) if (strcmp(ptr->exe, exe) == 0) return 0;
-	if ((saved_exe = SaveName(exe)) == NULL || (new_entry = (POLICY_MANAGER_ENTRY *) alloc_element(sizeof(POLICY_MANAGER_ENTRY))) == NULL) return -ENOMEM;
-	new_entry->exe = saved_exe;
-	/***** CRITICAL SECTION START *****/
-	spin_lock(&lock);
-	for (ptr = &policy_manager_list; ptr->next; ptr = ptr->next); ptr->next = new_entry;
-	spin_unlock(&lock);
-	/***** CRITICAL SECTION END *****/
-	return 0;
+	if (strncmp(data, KEYWORD_DELETE, KEYWORD_DELETE_LEN) == 0) {
+		data += KEYWORD_DELETE_LEN;
+		is_delete = 1;
+	}
+	return AddManagerEntry(data, is_delete);
 }
 
 static int ReadManagerPolicy(IO_BUFFER *head)
@@ -582,10 +681,10 @@ static int ReadManagerPolicy(IO_BUFFER *head)
 	if (!head->read_eof) {
 		POLICY_MANAGER_ENTRY *ptr = (POLICY_MANAGER_ENTRY *) head->read_var2;
 		if (!isRoot()) return -EPERM;
-		if (!ptr) ptr = policy_manager_list.next;
+		if (!ptr) ptr = policy_manager_list;
 		while (ptr) {
 			head->read_var2 = (void *) ptr;
-			if (io_printf(head, "%s\n", ptr->exe)) break;
+			if (!ptr->is_deleted && io_printf(head, "%s\n", ptr->manager->name)) break;
 			ptr = ptr->next;
 		}
 		if (!ptr) head->read_eof = 1;
@@ -598,10 +697,14 @@ static int IsPolicyManager(void)
 {
 	POLICY_MANAGER_ENTRY *ptr;
 	const char *exe;
+	const struct path_info *domainname = current->domain_info->domainname;
 	if (!sbin_init_started) return 1;
+	for (ptr = policy_manager_list; ptr; ptr = ptr->next) {
+		if (!ptr->is_deleted && ptr->is_domain && !pathcmp(domainname, ptr->manager)) return 1;
+	}
 	if ((exe = GetEXE()) == NULL) return 0;
-	for (ptr = policy_manager_list.next; ptr; ptr = ptr->next) {
-		if (strcmp(exe, ptr->exe) == 0) break;
+	for (ptr = policy_manager_list; ptr; ptr = ptr->next) {
+		if (!ptr->is_deleted && !ptr->is_domain && !strcmp(exe, ptr->manager->name)) break;
 	}
 	if (!ptr) { /* Reduce error messages. */
 		static pid_t last_pid = 0;
@@ -623,7 +726,7 @@ static int AddDomainPolicy(IO_BUFFER *head)
 {
 	char *data = head->write_buf;
 	struct domain_info *domain = head->write_var1;
-	int is_delete = 0, is_select = 0;
+	int is_delete = 0, is_select = 0, is_undelete = 0;
 	unsigned int profile;
 	if (!isRoot()) return -EPERM;
 	if (strncmp(data, KEYWORD_DELETE, KEYWORD_DELETE_LEN) == 0) {
@@ -632,6 +735,9 @@ static int AddDomainPolicy(IO_BUFFER *head)
 	} else if (strncmp(data, KEYWORD_SELECT, KEYWORD_SELECT_LEN) == 0) {
 		data += KEYWORD_SELECT_LEN;
 		is_select = 1;
+	} else if (strncmp(data, KEYWORD_UNDELETE, KEYWORD_UNDELETE_LEN) == 0) {
+		data += KEYWORD_UNDELETE_LEN;
+		is_undelete = 1;
 	}
 	UpdateCounter(CCS_UPDATES_COUNTER_DOMAIN_POLICY);
 	if (IsDomainDef(data)) {
@@ -640,6 +746,8 @@ static int AddDomainPolicy(IO_BUFFER *head)
 			domain = NULL;
 		} else if (is_select) {
 			domain = FindDomain(data);
+		} else if (is_undelete) {
+			domain = UndeleteDomain(data);
 		} else {
 			domain = FindOrAssignNewDomain(data, 0);
 		}
@@ -653,11 +761,6 @@ static int AddDomainPolicy(IO_BUFFER *head)
 	} else if (strncmp(data, KEYWORD_ALLOW_CAPABILITY, KEYWORD_ALLOW_CAPABILITY_LEN) == 0) {
 		return AddCapabilityPolicy(data + KEYWORD_ALLOW_CAPABILITY_LEN, domain, is_delete);
 #endif
-#ifdef CONFIG_TOMOYO_MAC_FOR_NETWORKPORT
-	} else if (strncmp(data, KEYWORD_ALLOW_BIND, KEYWORD_ALLOW_BIND_LEN) == 0 ||
-			   strncmp(data, KEYWORD_ALLOW_CONNECT, KEYWORD_ALLOW_CONNECT_LEN) == 0) {
-		return AddPortPolicy(data, domain, is_delete);
-#endif
 #ifdef CONFIG_TOMOYO_MAC_FOR_NETWORK
 	} else if (strncmp(data, KEYWORD_ALLOW_NETWORK, KEYWORD_ALLOW_NETWORK_LEN) == 0) {
 		return AddNetworkPolicy(data + KEYWORD_ALLOW_NETWORK_LEN, domain, is_delete);
@@ -692,113 +795,81 @@ static int ReadDomainPolicy(IO_BUFFER *head)
 		if (!isRoot()) return -EPERM;
 		for (domain = &KERNEL_DOMAIN; domain; domain = domain->next) {
 			struct acl_info *ptr;
-			if (GetDomainAttribute(domain) & DOMAIN_ATTRIBUTE_DELETED) continue;
+			if (domain->is_deleted) continue;
 			head->read_var1 = domain;
 			head->read_var2 = NULL; head->read_step = 1;
 		step1:
-			if (io_printf(head, "%s\n" KEYWORD_USE_PROFILE "%u\n\n", domain->domainname, domain->profile)) break;
+			if (io_printf(head, "%s\n" KEYWORD_USE_PROFILE "%u\n%s\n", domain->domainname->name, domain->profile, domain->quota_warned ? "quota_exceeded\n" : "")) break;
 			head->read_var2 = (void *) domain->first_acl_ptr; head->read_step = 2;
 		step2:
 			for (ptr = (struct acl_info *) head->read_var2; ptr; ptr = ptr->next) {
-				const unsigned int acl_type = GET_ACL_TYPE(ptr->type_hash);
+				const u8 acl_type = ptr->type;
 				const int pos = head->read_avail;
 				head->read_var2 = (void *) ptr;
+				if (ptr->is_deleted) continue;
 				if (0) {
 #ifdef CONFIG_TOMOYO_MAC_FOR_FILE
 				} else if (acl_type == TYPE_FILE_ACL) {
-					if (io_printf(head, "%d %s", ((FILE_ACL_RECORD *) ptr)->perm, ((FILE_ACL_RECORD *) ptr)->filename)
+					const unsigned char b = ptr->u.b[1];
+					if (io_printf(head, "%d %s%s", ptr->u.b[0], b ? "@" : "", b ? ((FILE_ACL_RECORD *) ptr)->u.group->group_name->name : ((FILE_ACL_RECORD *) ptr)->u.filename->name)
 						|| DumpCondition(head, ptr->cond)) {
 						head->read_avail = pos; break;
 					}
 #endif
 #ifdef CONFIG_TOMOYO_MAC_FOR_ARGV0
 				} else if (acl_type == TYPE_ARGV0_ACL) {
-					if (io_printf(head, KEYWORD_ALLOW_ARGV0 "%s %s", ((ARGV0_ACL_RECORD *) ptr)->filename, ((ARGV0_ACL_RECORD *) ptr)->argv0) ||
+					if (io_printf(head, KEYWORD_ALLOW_ARGV0 "%s %s", ((ARGV0_ACL_RECORD *) ptr)->filename->name, ((ARGV0_ACL_RECORD *) ptr)->argv0->name) ||
 						DumpCondition(head, ptr->cond)) {
 						head->read_avail = pos; break;
 					}
 #endif
 #ifdef CONFIG_TOMOYO_MAC_FOR_CAPABILITY
 				} else if (acl_type == TYPE_CAPABILITY_ACL) {
-					if (io_printf(head, KEYWORD_ALLOW_CAPABILITY "%s", capability2keyword(GET_ACL_HASH(ptr->type_hash))) ||
+					if (io_printf(head, KEYWORD_ALLOW_CAPABILITY "%s", capability2keyword(ptr->u.w)) ||
 						DumpCondition(head, ptr->cond)) {
 						head->read_avail = pos; break;
 					}
 #endif
-#ifdef CONFIG_TOMOYO_MAC_FOR_NETWORKPORT
-				} else if (acl_type == TYPE_BIND_ACL || acl_type == TYPE_CONNECT_ACL) {
-					const int is_stream = GET_ACL_HASH(ptr->type_hash);
-					const u16 min_port = ((NETWORK_ACL_RECORD *) ptr)->min_port, max_port = ((NETWORK_ACL_RECORD *) ptr)->max_port;
-					if (min_port != max_port) {
-						if (io_printf(head, "%s%s/%u-%u", acl_type == TYPE_CONNECT_ACL ? KEYWORD_ALLOW_CONNECT : KEYWORD_ALLOW_BIND, is_stream ? "TCP" : "UDP", min_port, max_port) ||
-							DumpCondition(head, ptr->cond)) {
-							head->read_avail = pos; break;
-						}
-					} else {
-						if (io_printf(head, "%s%s/%u", acl_type == TYPE_CONNECT_ACL ? KEYWORD_ALLOW_CONNECT : KEYWORD_ALLOW_BIND, is_stream ? "TCP" : "UDP", min_port) ||
-							DumpCondition(head, ptr->cond)) {
-							head->read_avail = pos; break;
-						}
-					}
-#endif
 #ifdef CONFIG_TOMOYO_MAC_FOR_NETWORK
-				} else if (acl_type == TYPE_IPv4_NETWORK_ACL) {
-					const char *keyword = network2keyword(GET_ACL_HASH(ptr->type_hash));
-					const u32 min_address = ((IPv4_NETWORK_ACL_RECORD *) ptr)->min_address, max_address = ((IPv4_NETWORK_ACL_RECORD *) ptr)->max_address;
-					const u16 min_port = ((IPv4_NETWORK_ACL_RECORD *) ptr)->min_port, max_port = ((IPv4_NETWORK_ACL_RECORD *) ptr)->max_port;
-					if (min_address != max_address) {
-						if (min_port != max_port) {
-							if (io_printf(head, KEYWORD_ALLOW_NETWORK "%s %u.%u.%u.%u-%u.%u.%u.%u %u-%u", keyword, HIPQUAD(min_address), HIPQUAD(max_address), min_port, max_port) || DumpCondition(head, ptr->cond)) {
-								head->read_avail = pos; break;
-							}
-						} else {
-							if (io_printf(head, KEYWORD_ALLOW_NETWORK "%s %u.%u.%u.%u-%u.%u.%u.%u %u", keyword, HIPQUAD(min_address), HIPQUAD(max_address), min_port) || DumpCondition(head, ptr->cond)) {
-								head->read_avail = pos; break;
-							}
+				} else if (acl_type == TYPE_IP_NETWORK_ACL) {
+					if (io_printf(head, KEYWORD_ALLOW_NETWORK "%s ", network2keyword(ptr->u.b[0]))) break;
+					switch (ptr->u.b[1]) {
+					case IP_RECORD_TYPE_ADDRESS_GROUP:
+						if (io_printf(head, "@%s", ((IP_NETWORK_ACL_RECORD *) ptr)->u.group->group_name->name)) goto print_ip_record_out;
+						break;
+					case IP_RECORD_TYPE_IPv4:
+						{
+							const u32 min_address = ((IP_NETWORK_ACL_RECORD *) ptr)->u.ipv4.min, max_address = ((IP_NETWORK_ACL_RECORD *) ptr)->u.ipv4.max;
+							if (io_printf(head, "%u.%u.%u.%u", HIPQUAD(min_address))) goto print_ip_record_out;
+							if (min_address != max_address && io_printf(head, "-%u.%u.%u.%u", HIPQUAD(max_address))) goto print_ip_record_out;
 						}
-					} else {
-						if (min_port != max_port) {
-							if (io_printf(head, KEYWORD_ALLOW_NETWORK "%s %u.%u.%u.%u %u-%u", keyword, HIPQUAD(min_address), min_port, max_port) || DumpCondition(head, ptr->cond)) {
-								head->read_avail = pos; break;
-							}
-						} else {
-							if (io_printf(head, KEYWORD_ALLOW_NETWORK "%s %u.%u.%u.%u %u", keyword, HIPQUAD(min_address), min_port) || DumpCondition(head, ptr->cond)) {
-								head->read_avail = pos; break;
+						break;
+					case IP_RECORD_TYPE_IPv6:
+						{
+							char buf[64];
+							const u16 *min_address = ((IP_NETWORK_ACL_RECORD *) ptr)->u.ipv6.min, *max_address = ((IP_NETWORK_ACL_RECORD *) ptr)->u.ipv6.max;
+							print_ipv6(buf, sizeof(buf), min_address);
+							if (io_printf(head, "%s", buf)) goto print_ip_record_out;
+							if (memcmp(min_address, max_address, 16)) {
+								print_ipv6(buf, sizeof(buf), max_address);
+								if (io_printf(head, "-%s", buf)) goto print_ip_record_out;
 							}
 						}
+						break;
 					}
-				} else if (acl_type == TYPE_IPv6_NETWORK_ACL) {
-					const char *keyword = network2keyword(GET_ACL_HASH(ptr->type_hash));
-					const u8 *min_address = ((IPv6_NETWORK_ACL_RECORD *) ptr)->min_address, *max_address = ((IPv6_NETWORK_ACL_RECORD *) ptr)->max_address;
-					const u16 min_port = ((IPv6_NETWORK_ACL_RECORD *) ptr)->min_port, max_port = ((IPv6_NETWORK_ACL_RECORD *) ptr)->max_port;
-					char buf1[64], buf2[64];
-					print_ipv6(buf1, sizeof(buf1), (const u16 *) min_address);
-					print_ipv6(buf2, sizeof(buf2), (const u16 *) max_address);
-					if (memcmp(min_address, max_address, 16)) {
-						if (min_port != max_port) {
-							if (io_printf(head, KEYWORD_ALLOW_NETWORK "%s %s-%s %u-%u", keyword, buf1, buf2, min_port, max_port)|| DumpCondition(head, ptr->cond)) {
-								head->read_avail = pos; break;
-							}
-						} else {
-							if (io_printf(head, KEYWORD_ALLOW_NETWORK "%s %s-%s %u", keyword, buf1, buf2, min_port) || DumpCondition(head, ptr->cond)) {
-								head->read_avail = pos; break;
-							}
-						}
-					} else {
-						if (min_port != max_port) {
-							if (io_printf(head, KEYWORD_ALLOW_NETWORK "%s %s %u-%u", keyword, buf1, min_port, max_port) || DumpCondition(head, ptr->cond)) {
-								head->read_avail = pos; break;
-							}
-						} else {
-							if (io_printf(head, KEYWORD_ALLOW_NETWORK "%s %s %u", keyword, buf1, min_port) || DumpCondition(head, ptr->cond)) {
-								head->read_avail = pos; break;
-							}
-						}
+					{
+						const u16 min_port = ((IP_NETWORK_ACL_RECORD *) ptr)->min_port, max_port = ((IP_NETWORK_ACL_RECORD *) ptr)->max_port;
+						if (io_printf(head, " %u", min_port)) goto print_ip_record_out;
+						if (min_port != max_port && io_printf(head, "-%u", max_port)) goto print_ip_record_out;
+					}
+					if (DumpCondition(head, ptr->cond)) {
+					print_ip_record_out: ;
+					head->read_avail = pos; break;
 					}
 #endif
 #ifdef CONFIG_TOMOYO_MAC_FOR_SIGNAL
 				} else if (acl_type == TYPE_SIGNAL_ACL) {
-					if (io_printf(head, KEYWORD_ALLOW_SIGNAL "%u %s", GET_ACL_HASH(ptr->type_hash), ((SIGNAL_ACL_RECORD *) ptr)->domainname) ||
+					if (io_printf(head, KEYWORD_ALLOW_SIGNAL "%u %s", ptr->u.w, ((SIGNAL_ACL_RECORD *) ptr)->domainname->name) ||
 						DumpCondition(head, ptr->cond)) {
 						head->read_avail = pos; break;
 					}
@@ -808,12 +879,14 @@ static int ReadDomainPolicy(IO_BUFFER *head)
 					const char *keyword = acltype2keyword(acl_type);
 					if (keyword) {
 						if (acltype2paths(acl_type) == 2) {
-							if (io_printf(head, "allow_%s %s %s", keyword, ((DOUBLE_ACL_RECORD *) ptr)->filename1, ((DOUBLE_ACL_RECORD *) ptr)->filename2)
+							const u8 b0 = ptr->u.b[0], b1 = ptr->u.b[1];
+							if (io_printf(head, "allow_%s %s%s %s%s", keyword, b0 ? "@" : "", b0 ? ((DOUBLE_ACL_RECORD *) ptr)->u1.group1->group_name->name : ((DOUBLE_ACL_RECORD *) ptr)->u1.filename1->name, b1 ? "@" : "", b1 ? ((DOUBLE_ACL_RECORD *) ptr)->u2.group2->group_name->name : ((DOUBLE_ACL_RECORD *) ptr)->u2.filename2->name)
 								|| DumpCondition(head, ptr->cond)) {
 								head->read_avail = pos; break;
 							}
 						} else {
-							if (io_printf(head, "allow_%s %s", keyword, ((SINGLE_ACL_RECORD *) ptr)->filename)
+							const u8 b = ptr->u.b[0];
+							if (io_printf(head, "allow_%s %s%s", keyword, b ? "@" : "", b ? ((SINGLE_ACL_RECORD *) ptr)->u.group->group_name->name : ((SINGLE_ACL_RECORD *) ptr)->u.filename->name)
 								|| DumpCondition(head, ptr->cond)) {
 								head->read_avail = pos; break;
 							}
@@ -842,9 +915,9 @@ static int ReadDomainProfile(IO_BUFFER *head)
 		}
 		if (!isRoot()) return -EPERM;
 		for (domain = head->read_var1; domain; domain = domain->next) {
-			if (GetDomainAttribute(domain) & DOMAIN_ATTRIBUTE_DELETED) continue;
+			if (domain->is_deleted) continue;
 			head->read_var1 = domain;
-			if (io_printf(head, "%u %s\n", domain->profile, domain->domainname)) break;
+			if (io_printf(head, "%u %s\n", domain->profile, domain->domainname->name)) break;
 		}
 		if (!domain) head->read_eof = 1;
 	}
@@ -870,7 +943,7 @@ static int ReadPID(IO_BUFFER *head)
 		if (p) domain = p->domain_info;
 		read_unlock(&tasklist_lock);
 		/***** CRITICAL SECTION END *****/
-		if (domain) io_printf(head, "%d %u %s", pid, domain->profile, domain->domainname);
+		if (domain) io_printf(head, "%d %u %s", pid, domain->profile, domain->domainname->name);
 		head->read_eof = 1;
 	}
 	return 0;
@@ -908,23 +981,35 @@ static int AddExceptionPolicy(IO_BUFFER *head)
 		data += KEYWORD_DELETE_LEN;
 		is_delete = 1;
 	}
-	if (strncmp(data, KEYWORD_DOMAIN_KEEPER, KEYWORD_DOMAIN_KEEPER_LEN) == 0) {
-		return AddDomainKeeperPolicy(data + KEYWORD_DOMAIN_KEEPER_LEN, is_delete);
-#ifdef CONFIG_TOMOYO_MAC_FOR_FILE
-	} else if (strncmp(data, KEYWORD_ALLOW_READ, KEYWORD_ALLOW_READ_LEN) == 0) {
-		return AddGloballyReadablePolicy(data + KEYWORD_ALLOW_READ_LEN, is_delete);
-#endif
+	if (strncmp(data, KEYWORD_KEEP_DOMAIN, KEYWORD_KEEP_DOMAIN_LEN) == 0) {
+		return AddDomainKeeperPolicy(data + KEYWORD_KEEP_DOMAIN_LEN, 0, is_delete);
+	} else if (strncmp(data, KEYWORD_NO_KEEP_DOMAIN, KEYWORD_NO_KEEP_DOMAIN_LEN) == 0) {
+		return AddDomainKeeperPolicy(data + KEYWORD_NO_KEEP_DOMAIN_LEN, 1, is_delete);
+	} else if (strncmp(data, KEYWORD_INITIALIZE_DOMAIN, KEYWORD_INITIALIZE_DOMAIN_LEN) == 0) {
+		return AddDomainInitializerPolicy(data + KEYWORD_INITIALIZE_DOMAIN_LEN, 0, is_delete, 0);
+	} else if (strncmp(data, KEYWORD_NO_INITIALIZE_DOMAIN, KEYWORD_NO_INITIALIZE_DOMAIN_LEN) == 0) {
+		return AddDomainInitializerPolicy(data + KEYWORD_NO_INITIALIZE_DOMAIN_LEN, 1, is_delete, 0);
 	} else if (strncmp(data, KEYWORD_INITIALIZER, KEYWORD_INITIALIZER_LEN) == 0) {
-		return AddInitializerPolicy(data + KEYWORD_INITIALIZER_LEN, is_delete);
+		return AddDomainInitializerPolicy(data + KEYWORD_INITIALIZER_LEN, 0, is_delete, 1);
+	} else if (strncmp(data, KEYWORD_NO_INITIALIZER, KEYWORD_NO_INITIALIZER_LEN) == 0) {
+		return AddDomainInitializerPolicy(data + KEYWORD_NO_INITIALIZER_LEN, 1, is_delete, 1);
 	} else if (strncmp(data, KEYWORD_ALIAS, KEYWORD_ALIAS_LEN) == 0) {
 		return AddAliasPolicy(data + KEYWORD_ALIAS_LEN, is_delete);
 	} else if (strncmp(data, KEYWORD_AGGREGATOR, KEYWORD_AGGREGATOR_LEN) == 0) {
 		return AddAggregatorPolicy(data + KEYWORD_AGGREGATOR_LEN, is_delete);
 #ifdef CONFIG_TOMOYO_MAC_FOR_FILE
+	} else if (strncmp(data, KEYWORD_ALLOW_READ, KEYWORD_ALLOW_READ_LEN) == 0) {
+		return AddGloballyReadablePolicy(data + KEYWORD_ALLOW_READ_LEN, is_delete);
 	} else if (strncmp(data, KEYWORD_FILE_PATTERN, KEYWORD_FILE_PATTERN_LEN) == 0) {
 		return AddPatternPolicy(data + KEYWORD_FILE_PATTERN_LEN, is_delete);
+	} else if (strncmp(data, KEYWORD_PATH_GROUP, KEYWORD_PATH_GROUP_LEN) == 0) {
+		return AddGroupPolicy(data + KEYWORD_PATH_GROUP_LEN, is_delete);
 	} else if (strncmp(data, KEYWORD_DENY_REWRITE, KEYWORD_DENY_REWRITE_LEN) == 0) {
 		return AddNoRewritePolicy(data + KEYWORD_DENY_REWRITE_LEN, is_delete);
+#endif
+#ifdef CONFIG_TOMOYO_MAC_FOR_NETWORK
+	} else if (strncmp(data, KEYWORD_ADDRESS_GROUP, KEYWORD_ADDRESS_GROUP_LEN) == 0) {
+		return AddAddressGroupPolicy(data + KEYWORD_ADDRESS_GROUP_LEN, is_delete);
 #endif
 	}
 	return -EINVAL;
@@ -946,7 +1031,7 @@ static int ReadExceptionPolicy(IO_BUFFER *head)
 #endif
 			head->read_var2 = NULL; head->read_step = 3;
 		case 3:
-			if (ReadInitializerPolicy(head)) break;
+			if (ReadDomainInitializerPolicy(head)) break;
 			head->read_var2 = NULL; head->read_step = 4;
 		case 4:
 			if (ReadAliasPolicy(head)) break;
@@ -962,6 +1047,16 @@ static int ReadExceptionPolicy(IO_BUFFER *head)
 		case 7:
 #ifdef CONFIG_TOMOYO_MAC_FOR_FILE
 			if (ReadNoRewritePolicy(head)) break;
+#endif
+			head->read_var2 = NULL; head->read_step = 8;
+		case 8:
+#ifdef CONFIG_TOMOYO_MAC_FOR_FILE
+			if (ReadGroupPolicy(head)) break;
+#endif
+			head->read_var2 = NULL; head->read_step = 9;
+		case 9:
+#ifdef CONFIG_TOMOYO_MAC_FOR_NETWORK
+			if (ReadAddressGroupPolicy(head)) break;
 #endif
 			head->read_eof = 1;
 			break;
@@ -1000,6 +1095,10 @@ static int AddSystemPolicy(IO_BUFFER *head)
 	if (strncmp(data, KEYWORD_ALLOW_CHROOT, KEYWORD_ALLOW_CHROOT_LEN) == 0)
 		return AddChrootPolicy(data + KEYWORD_ALLOW_CHROOT_LEN, is_delete);
 #endif
+#ifdef CONFIG_SAKURA_RESTRICT_PIVOT_ROOT
+	if (strncmp(data, KEYWORD_ALLOW_PIVOT_ROOT, KEYWORD_ALLOW_PIVOT_ROOT_LEN) == 0)
+		return AddPivotRootPolicy(data + KEYWORD_ALLOW_PIVOT_ROOT_LEN, is_delete);
+#endif
 #ifdef CONFIG_SAKURA_RESTRICT_AUTOBIND
 	if (strncmp(data, KEYWORD_DENY_AUTOBIND, KEYWORD_DENY_AUTOBIND_LEN) == 0)
 		return AddReservedPortPolicy(data + KEYWORD_DENY_AUTOBIND_LEN, is_delete);
@@ -1030,6 +1129,11 @@ static int ReadSystemPolicy(IO_BUFFER *head)
 #endif
 			head->read_var2 = NULL; head->read_step = 4;
 		case 4:
+#ifdef CONFIG_SAKURA_RESTRICT_PIVOT_ROOT
+			if (ReadPivotRootPolicy(head)) break;
+#endif
+			head->read_var2 = NULL; head->read_step = 5;
+		case 5:
 #ifdef CONFIG_SAKURA_RESTRICT_AUTOBIND
 			if (ReadReservedPortPolicy(head)) break;
 #endif
@@ -1081,12 +1185,12 @@ void CCS_LoadPolicy(const char *filename)
 		}
 		path_release(&nd);
 	}
-	
+
 #ifdef CONFIG_SAKURA
-	printk("SAKURA: 1.3.1   2006/12/08\n");
+	printk("SAKURA: 1.4   2007/04/01\n");
 #endif
 #ifdef CONFIG_TOMOYO
-	printk("TOMOYO: 1.3.1   2006/12/08\n");
+	printk("TOMOYO: 1.4   2007/04/01\n");
 #endif
 	if (!profile_loaded) panic("No profiles loaded. Run policy loader using 'init=' option.\n");
 	printk("Mandatory Access Control activated.\n");
@@ -1096,7 +1200,7 @@ void CCS_LoadPolicy(const char *filename)
 		struct domain_info *domain;
 		for (domain = &KERNEL_DOMAIN; domain; domain = domain->next) {
 			const u8 profile = domain->profile;
-			if (!profile_ptr[profile]) panic("Profile %u (used by '%s') not defined.\n", profile, domain->domainname); 
+			if (!profile_ptr[profile]) panic("Profile %u (used by '%s') not defined.\n", profile, domain->domainname->name); 
 		}
 	}
 }
